Developer Documentation for UNH EXS Testing Framework
=====================================================

##  Introduction

The goal of this document is to define the layout, structure, and
functionality of the UNH EXS Testing Framework and provide a means for
UNH EXS developers to maintain and update the framework as well as add
additional test cases and functionality to the testing framework as a
whole.

##  Structure

The UNH EXS framework is structured into key **include** files,
**setup** files, and the individual **test case** files.
These will be expanded and enumerated on below.
Each test case utilizes the same framework defined by command files and
only the individual test cases will change within the **test case**
files.

### Include Files

These files contain content that will be necessary for each
**test case** file.
In particular they cover common functions, parameters, and structs
that contain test specific parameters.
While all tests for a particular UNH EXS function is defined in the same
file each execution of the test is run by the framework itself.
The framework handles initialization, execution, reporting, and
tear-down of each test.
The mechanisms and operations on how the framework operates will be
explained in further sections.

#### param.h

This file holds the param structure used to store and pass parameters
and other values between the framework and the test functions.
This includes normal data types as well as EXS specific structures.
This document does not contain an exhaustive list of the kinds of
parameters used in the framework, but listed below are some important or
otherwise noteworthy parameters essential to the framework.

The **param** structure is initialized before any tests are run and the
parameters are set with default values or with values specified by the
user.

| Name   | Description |
| ----   | ----------- |
| `func` | The name of the function being tested. Initialized  immediately after **param** is initialized. |
| `addr` | Specifies the server interface address to be used in the testing framework. It is defaulted to a hard-coded value and can be specified by the user from the command line. |
| `port` | Specifies the port number to be used in the testing framework. It is defaulted to a hard-coded value and can be specified by the user from the command line. |
| `test_name` | String usually associated with the function being tested as it appears in the code with the relevant parameters displayed to show the user what is being tested. |
| `test_value` | String that denotes the expected result of the test. Usually denotes the error code that is expected, or the “OK” result. |
| `test_number` | Integer that denotes the test number associated with the test that is used when a specific test is specified by the user. This value is also printed with every test output. |
| `iter` | Linked list node that is used to iterate through the various test cases for a particular function. |
| `flags` | UNH EXS flag options values that are passed to the individual test functions during testing. The value of flags will differ based on the operation mode of the framework (blocking vs. non-blocking). |

#### fault.h

This file defines a function to set up a **fault handler** in order to
catch various test-terminating events such as SIGSEV, SIGABRT, and
SIGALRM.
This allows the test cases to terminate gracefully and report their
result in the final output.
Additionally, SIGALRM is used to safeguard against tests that do not
halt.
SIGALRM will be raised after a fixed amount of time for each test and
the result of such a signal will be reported in the final output.

The implementation of the fault handler is in [fault.c](tests/fault.c).
Note that when making changes to the fault handler, be careful not to
make calls to async-signal-unsafe functions.

#### print.h

This file contains various functions to aid in printing results (and
other messages) to stdout and the log file.
`print_error` and `print_message` are simple helper methods while
`print_macro` and the other `_macro` functions handle printing the full
results when a test is terminated.

##### print_error

```c
static int print_error(char* name);
```

This method prints the following string to stdout, where `%s` refers to
**name**:

```c
"A correct call to %s failed. The test suite will not terminate.\n"
```

##### print_mesasge

```c
static int print_message(char* msg);
```

This method prints **msg** to **stdout** if the **VERBOSE** flag is set
to **1** and prints msg to the **log** file regardless.

##### print_macro

```c
static int print_macro(int test, char* exp, char* name, int ret_value,
			int ret_expected, int ret_errno);
```

This method takes several parameters and prints them to **stdout** if the
**VERBOSE** flag is set to **1** and prints **msg** to the **log** file
regardless.
The mode of operation (**blocking** or **non-blocking**) is also
printed.
See [Results](#results).

##### print_mhandle_macro

```c
static int print_mhandle_macro(int test, char* exp, char* name,
			exs_mhandle_t ret_value, int ret_errno);
```

This method operates in the same way as `print_macro` except that
`ret_value` is of the `exs_mhandle_t` type.

##### print_qhandle_macro

```c
static int print_qhandle_macro(int test, char* exp, char* name,
			exs_qhandle_t ret_value, int ret_errno);
```

This method operates in the same way as `print_macro` except that
`ret_value` is of the `exs_qhandle_t` type.

#### cmdline.h

This file defines, maintains, and processes the various command line
options available to the user.
`cmdline.h` also initializes the **param** struct to be used in each
test case file.
This file also builds the linked list elements needed to specify which
test numbers will be run.
The details on the various command line options can be found in
[the UNH EXS User Documentation](tests/README.md).

##### setup_options

This method handles the command line options that the user may or may
not specify as well as initializing parameters in the **param** struct.
In particular, by default, the parameters are initialized to values that
will set the framework to run in **blocking mode**.
Note that by default, the **address** and **port** values are hardcoded
to values---they must be edited for the specific machine the framework
is running on, or specified in the command line every time.

Other than the specified command line options, the user can specify
tests that need to be run either **individually** or in a **range**.
These values are then parsed by `setup_options` and a linked list is
created which is then used when the test is executed.

#### result.h
This file contains most of the common includes, variables, and method
definitions that are used by all of the test files.
For example, it includes the other include files mentioned above, plus
the following methods:

```c
run_test(int i, void *test_func);
```

This method will run a test specified by i which is defined by the
`test_func` pointer to be a general way of running an individual test.
This method will create a child process to run the test and the
**param** struct will be passed to the new thread.
This method all sets up the SIGALRM that is used to terminate the
process started by `run_test` if it runs too long.

```c
list_tests(param *local_p, int i);
```

This method is a generalized mapping of a function to print out the
tests available for each function to the command line.
Each test defines its own method in its Setup Files.

```c
execute(int argc, char *argv[], void* test_func, int num_tests);
```

This method is what is passed to `run_test` and defines the tests that
need to be run when the user enters a command.
It initializes the **param** struct, calls `setup_options`, then either
lists the tests specified, or runs all tests if no test number is
specified, or runs the specified tests in the order they were received.

This file defines a global variable called `test_mode`, which specifies
if the framework is running in **blocking (synchronous)** or
**non-blocking (asynchronous)** mode.  This is needed so that
`print_macro` and friends can access this without being passed the
**param** struct.


### Setup Files

These files define reference functions to be used throughout the test
framework as calls to the functions with valid parameters that should
return success if the implementation the framework is run against is
functioning properly.
These are referred to as **OK** functions.
These functions are also used to set variables in the param struct that
other functions will use.
For example, `ok_exs_socket` will set **fd** to a valid socket address.

These files defines an array of test names that are used in the
`print_macro` call.
These files define the `list_function_tests` that will print out the
**test numbers** as well as **test names**.

#### OK Functions

**OK** test functions are functions that call a function with valid
parameters in both **blocking** and **non-blocking** mode (when
applicable).
These functions return success and failure, but do not report their own
results.
**OK** test functions are used as prerequisites for other tests.
If an **OK** test function fails, a specific error message will be
reported.
In general, if an **OK** function fails the current test should
terminate (or will fail due to parameters that were meant to be
initialized were not).

### Test Files

Test files must follow the following format:

```c
void *test_FUNCTION(void *test_void_ptr){
	param *p = (param *)test_void_ptr;
	int test = p->test_number;

	if( test == 1 ) {
		p->test_value = "ENOTCONN";
		p->test_name = all_FUNCTION_tests[test-1];
		setup_fault_catcher(p, 0);

		/** test code goes here **/
	} else if( test != 0 ) {
		/** test number not valid **/
		fprintf(stdout,"Invalid Test Number: %d\n", test);
	}
}

int main(int argc, char *argv[]){
	execute(argc, argv, test_send, FUNCTION_TESTS);
}
```

Note: **FUNCTION** should be replaced with the name of the applicable
function (excluding EXS)

Note that **test** is initialized to the `test_number` value in the
**param** struct that was initialized before the test was actually
called.
Each test will be defined in its own block and the details of the code
necessary for a test to complete will be expanded upon later in this
document.

Also note that if your test has a client and server component, it will
need to fork, and `setup_fault_catcher` should be called in the parent
and give the PID of the child process.
However, `setup_fault_catcher` must only be called once the `test_name`
and `test_value` functions have been called.

Each test also includes a call to the **OK** function call (defined in
the **setup** file related to the test files), usually at the end of the
test file.
This test will be a call with valid parameters in both **blocking** and
**non-blocking** mode (when applicable).

### Test Blocks

This section will attempt to describe the necessary elements to a test
block in the testing framework.
It will describe key functions that must be included, as well as how to
utilize the **param** struct, **blocking** versus **non-blocking** mode,
and using the `print_macro` properly.

#### Initialization

Two **param** struct variables must be set at the beginning of each test
block, `test_value` and `test_name`.
In most cases, `test_value` will be a string that represents the error
code that the test will return or the name of the **OK** function.
`test_name` should be set to the appropriate string in the array of
names in the **setup files** (indexed by the **test number**).

#### OK Functions

Each test (except `exs_init`) has prerequisite functions that need to be
used to set up the test case.
In each case, call these functions in the following manner:

```c
if( ok_exs_init(p) < 0 )
	exit(EXIT_FAILURE);
```

This ensures the prerequisite functions will terminate the test suite if
they fail.
If the prerequisite functions set or return any variables that are
necessary in the test block, retrieve them through the **param** struct.

#### Using Param Struct

The param struct is passed to each test block through the variable p.
This is a direct pointer to a **param** struct and thus the appropriate
variables can be accessed directly.
All parameters that will be passed to the function being tested should
be retrieved from **p**.

#### Use of FIFOs

This section only applies to **blocking** mode tests.
In some test cases it may be necessary to synchronize the execution of
the test---i.e. setting up the server for listening before a client
connects.
This is done using **FIFOs** which are created when the `setup_options`
function executes.
Here is an example on how to use the **FIFOs**:

Thread 1:
```c
int fd = open(exs_test_fifo_1,O_RDONLY);
/** executing second **/
close(fd);
/** execute after server **/
fd = open(exs_test_fifo_2, O_WRONLY);
close(fd);
```

Thread 2:
```c
/** executing first **/
int fd = open(exs_test_fifo_1,O_RDONLY);
close(fd);
/** execute after client **/
fd = open(exs_test_fifo_2, O_WRONLY);
close(fd);
```

#### Function Call

The function to be tested should always be isolated on its own line and
called using the appropriate exs function call.
The **return value** of the function should always be stored in an
**instance variable** and passed to the `print_macro` function to ensure
the result of the test is accurate.
For example:

```c
int ret = exs_send(p->fd, p->send_buf, p->SEND_BUFSIZE, p->flags,
			p->qh, p->ah, p->mh);
```

Note how the parameters passed to the function are all obtained from **p**.
This is best practice.

#### Using print_macro

The call to [`print_macro`](#print_macro) is crucial to ensure the
framework reports accurate results.
The return value of the function call to be tested should always
immediately be passed to `print_macro`.

#### Operation Mode

Some tests may behave differently or have different initialization steps
when operating in **non-blocking** rather than **blocking** mode.
For instance, `exs_qcreate` must be called in **non-blocking** mode but
is unnecessary in **blocking** mode.
To determine the current mode of operation, check the value in **p**
named **mode**.
If this value is *1* then the framework is operating in **non-blocking**
mode.
Otherwise, it is **blocking** mode.
If a test is not applicable in a given mode, simply include the
following line into the test block:

```c
if( test_mode == 1 )
	return 0;
```

This terminates test block if framework is in non-blocking mode.

#### Test Block Example
This is an example of a test block.
This test is for `exs_send` and is testing whether or not the function
will return the **ENOTCONN** error code if the **socket** specified has
never been connected.

```c
/* ENOTCONN: socket never connected */
if( test == 1 ){
	p->test_value = "ENOTCONN";
	p->test_name = all_send_tests[test-1];

	if (ok_exs_init(p) < 0)
		exit(EXIT_FAILURE);

	if (test_mode == 1)
		if (ok_exs_qcreate(p) < 0)
			exit(EXIT_FAILURE);

	if(ok_exs_socket(p) < 0)
		exit(EXIT_FAILURE);

	int ret = exs_send(p->fd, p->send_buf, p->SEND_BUFSIZE,
		p->flags, p->qh, p->ah, p->mh);

	print_macro(test, p->test_value, p->test_name, ret, -1,
		ENOTCONN);
}
```

##  Results

Test results will be printed by `print_macro` (and its variants) to
**stdout** and also to a **log** file.
A result will be printed on a single line in 7 separate parts, delimited
by a **":"**.
It will be sufficient to **diff** two result outputs to determine which
tests changed (provided the same tests were run in the same order).

| Field | Description |
| ----- | ----------- |
| Test Number | This is the number used to specify the test. Specifying this number in the command line will run this test. |
| Test Value | This is the **expected** test value to be returned from the test. Either an error code, or an **OK** function name. |
| Test Name | This displays the parameters of the call, along with additional information to differentiate the tests.|
| Mode | This displays either **blocking** or **non-blocking** mode. |
| Result | This simply displays **PASS** or **FAIL**. |
| Return Value | This is the value returned by the function, usually an error code or 0 on success. |
| Reason | This is the error code translated by **strerror**, or displays **SEGFAULT**, or **Time Out**. |

##  Modifying Existing Test Cases

To modify an **existing** test case, it is usually only necessary to edit
the **test file** and/or **setup file**.

To modify an existing test, the appropriate test block needs to be modified.
Be sure to modify the `print_macro` call if the **result** changes.

##  Adding Additional Test Cases

To add an additional test case to an **existing** function, it is usually
only necessary to modify the **test file** and/or **setup file**.

Adding a new test block requires incrementing the `FUNCTION_TESTS`
definition in the **setup file** and implementing a new test following
the guidelines in **Test Blocks**.

To add a new function to the framework, creating a new test file and
setup file is necessary, along with modifying `list_tests` in `result.h`.
It is critical to utilize the `print_macro` function correctly to ensure
results are properly displayed.

If a new function requires creating new function in `exs.c`, see
[Integration into UNH EXS Build System](#integration-into-unh-exs-build-system).


## Integration into UNH EXS Build System

The testing framework has been integrated into the UNH EXS build system.

### Framework Location

The testing framework is located in the [tests](tests) directory in the base
directory of the UNH EXS install.

### Building the Framework

The framework is not built during `make all install`.
Rather, the framework is only built if `make check` or `make distcheck`
is run.
This will compile all of the test cases in tests and run
[`test_all_check.sh`](#test-all-check-sh).
To run the tests manually, there are two options:

The first is to compile a test individually. For example:

```shell_session
$ make tests/test_exs_init
```

will build `test_exs_init` and it can be run as normal.

The second is to compile all tests but not run them.
To do this, use the command:

```shell_session
$ make TESTS= check
```

The output of test_all.sh is found in ./test-suite.log.

### Changing UNH EXS

Brand new functions that are added to UNH EXS need to be forward
declared in several locations in order to be built properly.
The changes need to be made in the following locations:

#### exs.h

`exs.h` is located at `include/exs.h`. Find the following section:

```c
/* Function Forward Declarations */
int		exs_init (int);
```

and add the new function signature here.

#### libexs.sym

`libexs.sym` is located at `libexs/libexs.sym`. The file starts with:

```
LIBEXS_1.2 {
global:
	exs_init;
	/* ... */
local:
	*;
};

LIBEXS_1.3 {
global:
	exs_getsockopt;
	/* ... */
local:
	*;
} LIBEXS_1.2;
```

This lists all public exported symbols in UNH EXS 1.2.0, and then the
public exported symbols in UNH EXS 1.3.0, which includes all symbols
exported in UNH EXS 1.2.0 (which makes UNH EXS backwards compatible).

Adding a new function generally requires adding a new block to this file
with the next version number, inheriting from the previous version.
The new function name needs to be added to this list (with a
semicolon).
If multiple functions are added in a single release, you merely need to
add the new function to the list.
Try to keep these lists sorted alphabetically.

#### Makefile.am

Two additions need to be made to Makefile.am.

The first is `EXTRA_tests_test_exs_init_SOURCES = \`.
Add the test case **setup file** to this list, e.g.
`tests/setup/exs_sendfile_setup.h`.

The second is a new `check_PROGRAMS` addition.
Add the following lines (with `exs_sendfile` swapped out for the exs
function you are adding:

```c
check_PROGRAMS += tests/test_exs_sendfile
tests_test_exs_sendfile_SOURCES = tests/test_exs_sendfile.c \
				tests/fault.c
```

#### test_all_check.sh

This script does some error checking and then runs the script
`test_all.sh` and pipes its output to grep which searches for the string
**"FAIL"**.
If it finds the string `test_all_check.sh` returns 1, else returns 0.
This is done to integrate into the existing tests in the UNH EXS build
system.
