User Documentation for UNH EXS Testing Framework
================================================

## Introduction

This document describes an non-exhaustive suite of tests that will run
nightly against the development installation of UNH EXS 1.4.0 to provide
a detailed report on the current status of the UNH EXS development and
support regression testing for future development.

The tests described in this document will include unit tests against
the **Extended Sockets API (ES-API)** specification published by the
**OpenGroup** as well as existing **man page** documentation included
with the 1.4.0 UNH EXS installation.

The end result of this test suite will issues reports detailing failing
tests. This mechanism is further described in **Framework**.

## Scope

This document will describe the layout, method, and reporting techniques
used to support regression testing for the UNH EXS 1.4.0 installation.
The main goal of this test suite is to directly verify the behavior of
UNH EXS functions to the the UNH EXS documentation as well as ES-API
conformance.

## Framework

Test cases will be structured in such a way that they contain
prerequisite tests that must pass in order to be run.
For example, if **Initialization** fails then the failure will be
reported followed by an error message stating that all further tests
will not be run because of the failure.
Test cases may be performed and reported on in a different order than
they appear on this document.
Test cases will be individually reported on in each group.
Test groups will maintain a running tally of successful test cases as
well as failure test cases.
Any failures will be displayed in-line with the rest of the testing,
unless such a failure requires the entire test group to not be tested.

An example test report is included at the end of this document.

A test result shall include **test group number**, **test number**,
**function name**, **return value**, **errno**, **result**, **message**
that can be independently verified.

The following contains an overview, in order, on how the tests are
structured such that each test succeeds a test that is required to pass
before it can be tested.

### Initialization

exs_init(), exs_socket()

### Client Connection

exs_bind()

#### Asynchronous

exs_connect()

#### Synchronous

exs_blocking_connect()

### Listening Post

exs_listen()

#### Asynchronous

exs_accept()

#### Synchronous

exs_blocking_accept()

### Memory Registration

exs_mderegister(), exs_mregister()

### Completion Queues

exs_qcreate(), exs_qdelete(), exs_qdequeue()

### Message Transfer

#### Asynchronous

exs_send(), exs_receive(), exs_sendfile()

#### Synchronous

exs_blocking_send(), exs_blocking_recv(), exs_read(), exs_write(), exs_blocking_sendfile()

### Socket Termination

exs_shutdown()

#### Asynchronous

exs_close()

#### Synchronous

exs_blocking_close()

### Other

exs_fcntl()

## Command Line Usage

The UNH EXS Testing Framework runs as a command line utility.
The results can be used to support Regression Testing, which will be
discussed in detail in the following subsections.
The options available to the user as well as detailed information for
the test cases that can be run can be viewed directly from the command
line.
Shell scripts are used to run the tests in aggregate.

### Command Line Options
The following options can be viewed from the command line by using the
**[-h]** or **[-H]** option on the command line at run time.

Command line options and their processing are defined in **cmdline.h**.

The options that will be displayed will be as follows:

| Option  | Description |
| ------- | ----------- |
| [-h -H] | Displays usage information. |
| [-l -L] | Displays all tests for a particular function, including the test number, test name, and expected return value. |
| [-t -T] | Displays an individual test for a particular function, including the test number, test name, and expected return value when **followed by a number**. |
| [-v -V] | Specifies the option for Verbose output – all passing and non passing tests will be printed to **stdout**. By default, passing results are not printed. |
| [-a -A] | Specifies the server interface address to be used in the testing framework. By default, this value is set to "localhost". |
| [-p -P] | Specifies the port number to be used in the testing framework. By default, this value is hard-coded in **port.h**. |
| [-n -N] | Specifies the option to run the testing framework in **non-blocking (asynchronous) mode**. By default, the testing framework runs in **blocking mode**. |
| [ 0 ]   | Specify which test to run. Can be separated by spaces. Invalid or numbers out of range of the test will not be accepted. If not test is selected, all tests will be run. |
| [0-1]   | Specify a range of tests to run. Invalid ranges will not be accepted. |

#### Individual Tests

Each EXS function can be run independently of each other.
To run a particular test it is necessary to compile the project and run
the appropriate object file with the appropriate command line options.
For example:

```
./test_exs_init -v
```

This will run all **exs_init** tests in **verbose** output mode.

#### Running All Tests

The UNH EXS Framework includes the bash script file **test_all.sh**.
This simple runs all test functions with the parameters to the script
passed to each function.

For example:

```
./test_all.sh -v -p 55555
```

This will run all tests in verbose mode with a port number of 55555.
Note, the project must be compiled before **test_all.sh** can be run.

### Regression Testing

The UNH EXS Framework reports to **stdout** as well as a **log** file.
Each line includes the test number, test name, expected result, observed
result, and test result.
It shall be sufficient to **diff** different outputs from running the
framework to determine if the results have changed and if so in what
way.
The mechanism to achieve true automated regression testing is not in the
scope of this document and is left as an exercise for the reader.

##  Test Groups
The tests described in this document are separated into various groups
to better segment the initialization required for each tests and to keep
results consistent across groups.

The groups include:

 * ES-API Conformance
 * EXS Conformance

###  ES-API Conformance

The goals of these tests are to verify that all implemented ES-API
functions correctly conform to specification published by the
**OpenGroup**.

These tests will attempt to:

 * Verify all functions return **0** on success cases.
 * Verify all functions return **-1** on failure cases.
 * Verify all functions set **errno** appropriately on failure cases.

Undefined behavior will be reported as failures.

#### exs_accept()

The following tests will exercise exs_accept() and its various success
and error cases.

**Prerequisites:** exs_init(), exs_socket(), exs_bind(), exs_listen(),
exs_connect(), exs_qcreate(), exs_qdequeue()

```c
int exs_accept(	int socket,
		exs_acceptaddr_t *addrvec,
		int addrvec_cnt,
		int flags,
		exs_qhandle_t qhandle);
```

| Parameter | Description |
| --------- | ----------- |
| `socket` | file descriptor used to specify the socket to be used |
| `addrvec` | pointer to array of exs_acceptaddr structures |
| `addrvec_cnt` | specifies the number of connections that should be accepted |
| `flags` | additional options to be specified (blocking/non-blocking) |
| `qhandle` | specifies the destination event queue |

##### Test [1]: EPERM
```c
exs_accept(p->fd, &acceptaddr, 1, EXS_BLOCK, NULL) [no exs_init]
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2]: EINVAL
```c
exs_accept(p->fd, &acceptaddr, 1, BAD_FLAG, NULL)
```

Verify the function returns -1 and sets errno to [EINVAL] if flag is invalid

##### Test [3]: EINVAL
```c
exs_accept(p->fd, &acceptaddr, NEGATIVE_ADDRVEC_CNT, 0, qh)
```

Verify the function returns -1 and sets errno to [EINVAL] if addrvec_cnt <= 0 when non blocking
In blocking mode, this test case is not performed.

##### Test [4]: EINVAL
```c
exs_accept(p->fd, &acceptaddr, ZERO_ADDRVEC_CNT, EXS_BLOCK, qh)
```

Verify the function returns -1 and sets errno to [EINVAL] if addrvec_cnt != 1 when blocking
In non-blocking mode, this test case is not performed.

##### Test [5]: EINVAL
```c
exs_accept(p->fd, &acceptaddr, 1, 0, INVALID_QHANDLE)
```

Verify the function returns -1 and sets errno to [EINVAL] if qhandle is not a valid event queue.

##### Test [6]: EINVAL
```c
exs_accept(p->fd, &acceptaddr, 0, EXS_BLOCK, NULL) [no exs_listen]
```

Verify the function returns -1 and sets errno to [EINVAL] if the socket
pointed to by fd is not accepting connections ( exs_listen() has not
been called)

##### Test [7]: EBADF
```c
exs_accept(BAD_FD, &acceptaddr, 1, EXS_BLOCK, NULL)
```

Verify the function returns -1 and sets errno to [EINVAL] if fd is invalid

##### Test [8]: OK_EXS_ACCEPT

```c
exs_accept(p->fd, &acceptaddr, 1, EXS_BLOCK, NULL)
```

Verify the function returns 0 if passed correct parameters.

#### exs_connect()

The following tests will exercise exs_connect() and its various success and error cases.

Prerequisites: exs_init(), exs_socket(), exs_bind(), exs_listen(), exs_qcreate(), exs_qdequeue()

```c
int exs_connect(	int sockno,
			const struct sockaddr *address,
			socklent_t address_len,
			int flags,
			struct timeval *timeout,
			exs_qhandle_t qhandle,
			exs_ahandle_t ahandle);
```

| Parameter | Description |
| --------- | ----------- |
| `sockno` | file descriptor used to specify the socket to be used |
| `address` | points to a sockaddr structure containing the peer address |
| `address_len` | specifies the length of the sockaddr structure pointed to by address |
| `flags` | additional options to be specified (blocking/non-blocking) |
| `timeout` | specifies how long the asynchronous connect operation should take before timing out |
| `qhandle` | specifies the destination event queue |
| `ahandle` | arbitrary pointer value chose by the user |

##### Test [1]: EPERM
```c
exs_connect(fd, ai_addr, ai_addrlen, flags, to, qh, ah) [no exs_init]
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2]: EINVAL
```c
exs_connect(fd, ai_addr, ai_addrlen, -1, to, qh, ah)
```

Verify the function returns -1 and sets errno to [EINVAL] if flags is invalid (i.e. set to -1).

##### Test [3]: EINVAL
```c
exs_connect(fd, ai_addr, ai_addrlen, BAD_FLAG, to, qh, ah)
```

Verify the function returns -1 and sets errno to [EINVAL] if flags is invalid.

```c
##### Test [4]: EBADF
```
exs_connect(BAD_FD, ai_addr, ai_addrlen, flags, to, qh, ah)

Verify the function returns -1 and sets errno to [EBADF] if the socket argument is not a valid file descriptor.


```c
##### Test [5]: EISCONN
```
exs_connect(CONNECTED_FD, ai_addr, ai_addrlen, flags, to, qh, ah)

Verify the function returns -1 and sets errno to [EISCONN] if the socket provided is already connected or listened to.

##### Test [6]: OK_EXS_CONNECT
```c
exs_connect(fd, ai_addr, ai_addrlen, flags, to, qh, ah)
```

Verify the function returns 0 if passed correct parameters.

#### exs_init()

The following tests will exercise exs_init() and its various success and error cases.

Prerequisites:

```c
int exs_init(int version);
```

| Parameter | Description |
| --------- | ----------- |
| `version` | requested version of the API to be initialized |

##### Test [1]: ENOTSUP
```c
exs_init(INVALID_VERSION)
```

Verify the function returns -1 and sets errno to [ENOTSUP] if the version parameter provided to exs_init() is not supported.

##### Test [2]: OK_EXS_INIT
```c
exs_init(VERSION)
```

Verify the function returns 0 if passed a correct version parameter.

##### Test [3]: EALREADY
```c
exs_init(VERSION) [exs_init already]
```

Verify the function returns -1 and sets errno to [EALREADY] if exs_init() has already been called in the same process.

#### exs_mderegister()

The following tests will exercise exs_mderegister() and its various success and error cases.

Prerequisites: exs_init()

```c
exs_mhandle_t exs_mderegister(exs_mhandle_t mhandle, int flags);
```

| Parameter | Description |
| --------- | ----------- |
| `mhandle` | specifies the handle to the memory being deregistered |
| `flags` | additional options to be specified |

##### Test [1]: EPERM
```c
exs_mderegister(mh, 0)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2]: EINVAL
```c
exs_mderegister(NULL, 0)
```

Verify the function returns -1 and sets errno to [EINVAL] if mhandle is NULL.

##### Test [3]: EINVAL
```c
exs_mderegister(EXS_MHANDLE_INVALID, 0)
```

Verify the function returns -1 and sets errno to [EINVAL] if mhandle is set to EXS_MHANDLE_INVALID.

##### Test [4]: EINVAL
```c
exs_mderegister(EXS_MHANDLE_UNREGISTERED, 0)
```

Verify the function returns -1 and sets errno to [EINVAL] if mhandle is set to EXS_MHANDLE_UNREGISTERED.

##### Test [5]: OK_EXS_MDEREGISTER
```c
exs_mderegister(mh, 0)
```

Verify the function returns 0 if passed correct parameters.

####  exs_mregister()

The following tests will exercise exs_mregister() and its various success and error cases.

Prerequisites: exs_init()

```c
exs_mhandle_t exs_mregister(void *addr, size_t size, int flags);
```

| Parameter | Description |
| --------- | ----------- |
| `addr` | address to the application memory to be registered |
| `size` | length of the application memory to be registered |
| `flags` | additional options to be specified |

##### Test [1]: EPERM
```c
exs_mregister(buf, BUFSIZE, 0)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [3]: EINVAL
```c
exs_mregister(buf, ZERO_SIZE, 0)
```

Verify the function returns EXS_MHANDLE_INVALID and sets errno to[EINVAL] if size is less than or equal to 0.

##### Test [4]: EINVAL
```c
exs_mregister(buf, BUFSIZE, BAD_FLAG)
```

Verify the function returns EXS_MHANDLE_INVALID and sets errno to[EINVAL] if flag is invalid.

##### Test [5]: EINVAL
```c
exs_mregister(NULL_ADDR, BUFSIZE, 0)
```

Verify the function returns EXS_MHANDLE_INVALID and sets errno to[EINVAL] if addr is NULL.

##### Test [6]: OK_EXS_MREGISTER
```c
exs_mregister(buf, BUFSIZE, 0)
```

Verify the function returns a valid opaque memory handle if passed correct parameters.



Unimplemented Tests:
Verify the function returns EXS_MHANDLE_INVALID and sets errno to[EACCES] if the function fails due to memory permissions.

####  exs_qcreate()

The following tests will exercise exs_qcreate() and its various success and error cases.

Prerequisites: exs_init()

```c
exs_qhandle_t exs_qcreate(int depth);
```

| Parameter | Description |
| --------- | ----------- |
| `depth` | specifies the guaranteed minimum number of events that can be stored in the queue. |

##### Test [1]: EPERM
```c
exs_qcreate(10)
```

Verify the function returns EXS_QHANDLE_INVALID and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [3]: OK_EXS_QCREATE
```c
exs_qcreate(10)
```

Verify the function returns a valid event queue if passed correct parameters.

Unimplemented Tests:
Verify the function returns EXS_QHANDLE_INVALID and sets errno to[EAGAIN] if the allocation of internal resources failed but a subsequent request may succeed.
Verify the function returns EXS_QHANDLE_INVALID and sets errno to[EINVAL] if the event queue resources were exceeded.

####  exs_qdelete()

The following tests will exercise exs_qdelete() and its various success and error cases.

Prerequisites: exs_init(), exs_qcreate()

```c
int exs_qdelete( exs_qhandle_t qhandle);
```

| Parameter | Description |
| --------- | ----------- |
| `qhandle` | specifies the event queue to be deleted |

##### Test [1]: EPERM
```c
exs_qdelete(qh)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [3]: EINVAL
```c
exs_qdelete(INVALID_QHANDLE)
```

Verify the function returns -1 and sets errno to [EINVAL] if qhandle is not a valid event queue.

##### Test [4]: OK_EXS_QDELETE
```c
exs_qdelete(qh)
```

Verify the function returns 0 if passed a valid qhandle.

Unimplemented Tests:
Verify the function returns -1 and sets errno to [EBUSY] if there are still asynchronous operations in progress for this event queue.

####  exs_qdequeue()

The following tests will exercise exs_qdequeue() and its various success and error cases.

Prerequisites: exs_init(), exs_qcreate()

```c
int exs_qdequeue(	exs_qhandle_t qhandle,
			exs_event_t *evtvec,
			int evtvec_cent,
			const struct timeval *timeout);
```

| Parameter | Description |
| --------- | ----------- |
| `qhandle` | specifies an event queue |
| `*evtvec` | specifies the address of an array of exs_event_t types |
| `evtvec_cnt` | specifies the number of exs_event_t elements in the evtvec array |
| `timeout` | specifies how long the call should wait before timing out |


##### Test [1]: EPERM
```c
exs_qdequeue(qh, evtvec, evtvec_cent,to)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2]: EINVAL
```c
exs_qdequeue(INVALID_QH, evtvec, evtvec_cent,to)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [3]: EINVAL
```c
exs_qdequeue(qh, evtvec, ZERO_EVTVEC_CENT,to)
```

Verify the function returns -1 and sets errno to [EINVAL] if evtvec_cnt is less than 1.

Unimplemented Tests:
Verify the function returnsthe correct number of dequeued events if passed correct parameters.
Verify the function returns -1 and sets errno to [EFAULT] if:
The address range specified by evtvec and evtvec_cnt is not valid.
timeout is invalid
Verify the function returns -1 and sets errno to [EINVAL] if evtvec_cnt is greater than EXS_EVTVEC_MAX.



####  exs_recv()

The following tests will exercise exs_recv() and its various success and error cases.

Prerequisites: exs_init(), exs_socket(), exs_bind(), exs_listen(), exs_accept()

```c
ssize_t exs_recv(	int socket,
			void *buffer,
			size_t length,
			int flags
			exs_qhandle_t qhandle,
			exs_ahandle_t ahandle,
			exs_mhandle_t mhandle);
```

| Parameter | Description |
| --------- | ----------- |
| `socket` | specifies the socket file descriptor |
| `buffer` | points to a buffer where the message should be stored |
| `length` | specifies the length in bytes of the buffer pointed to by the buffer argument |
| `flags` | specifies additional options (blocking/non-blocking) |
| `qhandle` | specifies an event queue |
| `ahandle` | specifies an arbitrary pointer value chosen by the user |
| `mhandle` | specifies a registered memory handle |

##### Test [1]: EPERM
```c
exs_recv(fd, recv_buf, RECV_BUFSIZE, flags, qh, ah, mh)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2]: EINVAL
```c
exs_recv(fd, recv_buf, RECV_BUFSIZE, flags, INVALID_QH, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EINVAL] if qhandle is invalid.

##### Test [3]: EINVAL
```c
exs_recv(fd, recv_buf, RECV_BUFSIZE, 0, INVALID_QH, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EINVAL] if qhandle is NULL and EXS_BLOCK is not specified.

##### Test [4]: EINVAL
```c
exs_recv(fd, recv_buf, ZERO_LENGTH, flags, qh, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EINVAL] if length is not at least 1.
##### Test [5]: EINVAL
```c
exs_recv(fd, recv_buf, MAX_LENGTH, flags, qh, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EINVAL] if length is greater than the maximum supported length for the RDMA channel.

##### Test [6]: EINVAL
```c
exs_recv(fd, BAD_BUFFER, RECV_BUFSIZE, flags, qh, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EINVAL] if buffer is NULL.

##### Test [7]: EOPNOTSUPP
```c
exs_recv(fd, recv_buf, RECV_BUFSIZE, BAD_FLAG, qh, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EOPNOTSUPP] if at least one of the flags passed to the function are not supported.

##### Test [8]: EBADF
```c
exs_recv(BAD_FD, recv_buf, RECV_BUFSIZE, flags, qh, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EBADF] if the function was passed a bad file descriptor.

##### Test [9]: ENOTCONN
```c
exs_recv(BAD_FD, recv_buf, RECV_BUFSIZE, flags, qh, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [ENOTCONN] if the socket was never connected to prior to the function call.

##### Test [10]: OK_EXS_RECV
```c
exs_recv(BAD_FD, recv_buf, RECV_BUFSIZE, flags, qh, ah, mh)
```

Verify the function returns 0 if passed all valid parameters.

#### exs_send()

The following tests will exercise exs_send() and its various success and error cases.

Prerequisites:  exs_init(), exs_socket(), exs_connect()

```c
ssize_t exs_send(	int socketno,
			const void *buffer,
			size_t length,
			int flags,
			exs_qhandle_t qhandle,
			exs_ahandle_t ahandle,
			exs_mhandle_t mhandle);
```

| Parameter | Description |
| --------- | ----------- |
| `socket` | specifies the socket file descriptor |
| `buffer` | points to a buffer containing the message to send |
| `length` | specifies the length in bytes of the buffer pointed to by the buffer argument |
| `flags` | specifies additional options (blocking/non-blocking) |
| `qhandle` | specifies an event queue |
| `ahandle` | specifies an arbitrary pointer value chosen by the user |
| `mhandle` | specifies a registered memory handle |

##### Test [1]: ENOTCONN
```c
exs_send(NOT_CONNECTED_FD, send_buf, SEND_BUFSIZE, flags, qh, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [ENOTCONN] if the socket was never connected.

##### Test [2]: EPERM
```c
exs_send(fd, send_buf, SEND_BUFSIZE, flags, qh, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [3]: EINVAL
```c
exs_send(fd, send_buf, SEND_BUFSIZE, flags, INVALID_QHANDLE, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EINVAL] if qhandle is invalid.

##### Test [4]: EINVAL
```c
exs_send(fd, send_buf, SEND_BUFSIZE, 0, INVALID_QHANDLE, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EINVAL] if qhandle is NULL and EXS_BLOCK is not specified and qhandle is NULL and EXS_UNSIGNALED is not specified.

##### Test [5]: EINVAL
```c
exs_send(BAD_BUFFER, send_buf, SEND_BUFSIZE, flags, qh, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EINVAL] if buffer is NULL.

##### Test [7]: EOPNOTSUPP
```c
exs_send(fd, send_buf, SEND_BUFSIZE, BAD_FLAG, qh, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EOPNOTSUPP] if flags is not supported.

##### Test [8]: EBADF
```c
exs_send(BAD_FD, send_buf, SEND_BUFSIZE, flags, qh, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EBADF] if the function was passed a bad file descriptor.

##### Test [10]: EPIPE
```c
exs_send(fd, send_buf, SEND_BUFSIZE, flags, qh, ah, mh)
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EPIPE] if the connection has been gracefully shutdown for writing.

##### Test [16]: OK_EXS_SEND
```c
exs_send(fd, send_buf, SEND_BUFSIZE, flags, qh, ah, mh)
```

Verify the function returns 0 if passed all valid parameters.

Unimplemented Tests:
Verify the function returns -1, no event is posted to qhandle, and sets errno to [ECONNRESET] if the connection has been terminated abruptly.
Verify the function returns -1, no event is posted to qhandle, and sets errno to [EINVAL] if EXS_UNSIGNALED and EXS_BLOCK were both specified.
Verify the function returns -1, no event is posted to qhandle, and sets errno to [EBUSY] if no send credits were available at the time of this function call

#### exs_sendfile()

The following tests will exercise exs_sendfile() and its various success and error cases.

Prerequisites:  exs_init(), exs_socket(), exs_connect()

```c
ssize_t exs_sendfile(	int socket,
			exs_xferfile_t *sendvec,
			int sendvec_cnt,
			int flags,
			exs_qhandle_t qhandle,
			exs_qhandle_t ahandle);
```

| Parameter | Description |
| --------- | ----------- |
| `socket` | specifies the socket file descriptor |
| `sendvec` | specifies an array of file descriptors and memory buffers |
| `sendvec_cnt` | specifies the number of elements in sendvec array |
| `flags` | specifies additional options (blocking/non-blocking) |
| `qhandle` | specifies an event queue |
| `ahandle` | specifies an arbitrary pointer value chosen by the user |

##### Test [1]: EPERM
```c
exs_sendfile(fd, sendvec, sendvec_cnt, flags, qh, ah);
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2]: ENOTCONN
```c
exs_sendfile(NOT_CONNECTED_FD, sendvec, sendvec_cnt, flags, qh, ah);
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [ENOTCONN] if the socket was never connected.

##### Test [3]: EINVAL
```c
exs_sendfile(BAD_FD, sendvec, sendvec_cnt, flags, qh, ah);
```

Verify the function returns -1 and sets errno to [EBADF] if fd is invalid.

##### Test [4]: ENODEV
```c
exs_sendfile(fd, sendvec, sendvec_cnt, flags, qh, ah);
```

Verify the function returns -1 and sets errno to [ENODEV] if the file descriptor cannot be mmapped.

##### Test [5]: EINVAL
```c
exs_sendfile(fd, sendvec, sendvec_cnt, flags, qh, ah);
```

Verify the function returns -1 and sets errno to [EINVAL] if exs_fileds is invalid.

##### Test [6]: EINVAL
```c
exs_sendfile(fd, sendvec, sendvec_cnt, flags, qh, ah);
```

Verify the function returns -1 and sets errno to [EINVAL] if exs_path is invalid.

##### Test [7]: EFBIG
```c
exs_sendfile(fd, sendvec, sendvec_cnt, flags, qh, ah);
```

Verify the function returns -1 and sets errno to [EFBIG] if file size > 1GB.

##### Test [8]: EINVAL
```c
exs_sendfile(fd, sendvec, sendvec_cnt, flags, qh, ah);
```

Verify the function returns -1 and sets errno to [EINVAL] if offset + length > fize size.

##### Test [9]: EINVAL
```c
exs_sendfile(fd, sendvec, sendvec_cnt, flags, qh, ah);
```

Verify the function returns -1 and sets errno to [EINVAL] if offset > file size.

##### Test [10]: EINVAL
```c
exs_sendfile(fd, sendvec, BAD_SENDVEC_CNT, flags, qh, ah);
```

Verify the function returns -1 and sets errno to [EINVAL] if sendvec_cnt is invalid.

##### Test [11]: EINVAL
```c
exs_sendfile(fd, sendvec, sendvec_cnt, BAD_FLAG, qh, ah);
```

Verify the function returns -1 and sets errno to [EINVAL] if flags is invalid.

##### Test [12]: EINVAL
```c
exs_sendfile(fd, sendvec, sendvec_cnt, EXS_UNSIGNALED & EXS_BLOCK, qh, ah);
```

Verify the function returns -1 and sets errno to [EINVAL] if EXS_UNSIGNALED and EXS_BLOCK were both specified.

##### Test [13]: EINVAL
```c
exs_sendfile(fd, sendvec, sendvec_cnt, EXS_BLOCK, INVALID_QHANDLE, ah);
```

Verify the function returns -1 and sets errno to [EINVAL] if EXS_BLOCK is specified and qhandle is invalid.

##### Test [14]: EPIPE
```c
exs_sendfile(fd, sendvec, sendvec_cnt, flags, qh, ah);
```

Verify the function returns -1 and sets errno to [EPIPE] if the socket is gracefully shut down for writing.

##### Test [15]: OK_EXS_SENDFILE
```c
exs_sendfile(fd, sendvec, sendvec_cnt, flags, qh, ah);
```

Verify the function returns 0 if passed all valid parameters.

##### Test [16]: OK_EXS_SENDFILE
```c
exs_sendfile(fd, sendvec, sendvec_cnt, flags, qh, ah);
```

Verify the function returns 0 if passed all valid parameters and exs_sendfile() is used to send the same file immediately afterward.

##### Test [15]: OK_EXS_SENDFILE
```c
exs_sendfile(fd, sendvec, sendvec_cnt, flags, qh, ah);
```

Verify the function returns 0 if passed all valid parameters using exs_pathvec rather than exs_fdvec.

###  EXS Conformance
The goals of these tests are to verify that all EXS functions that are
not part of the ES-API specification correctly conform to documentation
provided by the UNH EXS installation.

These tests will attempt to:

 * Verify all functions return 0 on success cases.
 * Verify all functions return -1 on failure cases.
 * Verify all functions set errno appropriately on failure cases.
 * Undefined behavior will be reported as failures.

####  exs_bind()

The following tests will exercise exs_bind() and its various success and error cases.

Prerequisites: exs_init(), exs_socket()

```c
int exs_bind(	int fd,
		struct sockaddr *address,
		socklen_t addrlen);
```

| Parameter | Description |
| --------- | ----------- |
| `fd` | specifies socket file descriptor |
| `address` | pointer to a structure of type struct sockaddr that contains port address and number information |
| `addrlen` | size in bytes of the structure pointed to by address |

##### Test [1]: EPERM
```c
exs_bind(fd, ai_addr, ai_addrlen)
```
Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2]: EBADF
```c
exs_bind(BAD_FD, ai_addr, ai_addrlen)
```

Verify the function returns -1 and sets errno to [EBADF] if fd is invalid.

##### Test [3]: EINVAL
```c
exs_bind(CONNECTED_FD, ai_addr, ai_addrlen)
```

Verify the function returns -1 and sets errno to [EINVAL] if the socket provided by fd has already been bound.

##### Test [4]: OK_EXS_BIND
```c
exs_bind(fd, ai_addr, ai_addrlen)
```

Verify the function returns 0 if passed correct parameters.

#### exs_blocking_accept()

The following tests will exercise exs_blocking_accept() and its various success and error cases.

Prerequisites: exs_init(), exs_socket(), exs_bind(), exs_listen()

```c
int exs_blocking_accept(	int sockno,
				struct sockaddr *addr,
				socklen_t *addrlen);
```

| Parameter | Description |
| --------- | ----------- |
| `sockno` | specifies socket file descriptor |
| `addr` | points to an array of exs_acceptaddr structures |
| `addrlen` | specifies the number of connections that should be accepted |

##### Test [1]: EPERM
```c
exs_blocking_accept(fd, addr_store, addrlen)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2]: EINVAL
```c
exs_blocking_accept(fd, addr_store, addrlen)
```

Verify the function returns -1 and sets errno to [EINVAL] if socket is not accepting connections ( exs_listen() has not been called).

##### Test [3]: OK_EXS_BLOCKING_ACCEPT
```c
exs_blocking_accept(fd, addr_store, addrlen)
```

Verify the function returns the socket for the new connection to the remote client if passed correct parameters.

Unimplemented Tests:
Verify the function returns -1 and sets errno to [EINVAL] if:
flags is not supported
addrvec_cnet is less than or equal to 0
qhandle is not a valid event queue
Verify the function returns -1 and sets errno to [EBADF] if the socket argument is not a valid file descriptor.

####  exs_blocking_close()

The following tests will exercise exs_blocking_close() and its various success and error cases.

Prerequisites: exs_init(), exs_socket()

```c
int exs_blocking_close(int sockno);
```

| Parameter | Description |
| --------- | ----------- |
| `sockno` | specifies socket file descriptor |

##### Test [1]: EPERM
```c
exs_blocking_close(fd)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2]: OK_EXS_BLOCKING_CLOSE
```c
exs_blocking_close(fd)
```

Verify the function returns 0 and correctly closes the socket if passed the correct parameters.

Unimplemented Tests:
Verify the function returns -1 and sets errno to [EINVAL]if:
flags are not supported
qhandle is invalid
Verify the function returns -1 and sets errno to [EBADF] if socket is invalid.

####  exs_blocking_connect()

The following tests will exercise exs_blocking_connect() and its various success and error cases.

Prerequisites: exs_init(), exs_socket()

```c
int exs_blocking_connect(	int sockno,
				const struct sockaddr *address,
				socklen_t address_len);
```

| Parameter | Description |
| --------- | ----------- |
| `sockno` | specifies socket file descriptor |
| `address` | points to a sockaddr structure containing the peer address |
| `address_len` | specifies the length of the sockaddr structure pointed to by the address argument |

##### Test [1]: EPERM
```c
exs_blocking_connect(fd, ai_addr, ai_addrlen, flags, mh, qh, ah)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2]: EISCONN
```c
exs_blocking_connect(CONNECTED_FD, ai_addr, ai_addrlen, flags, mh, qh, ah)
```

Verify the function returns -1 and sets errno to [EISCONN] if the socket provided is already connected or listened to.

##### Test [3]: OK_EXS_BLOCKING_CONNECT
```c
exs_blocking_connect(CONNECTED_FD, ai_addr, ai_addrlen, flags, mh, qh, ah)
```

Verify the function returns 0 if passed correct parameters.

####  exs_blocking_recv()

The following tests will exercise exs_blocking_recv() and its various success and error cases.

Prerequisites: exs_init(), exs_socket(), exs_bind(), exs_listen(), exs_accept()

```c
ssize_t exs_blocking_recv(	int sockno,
				void *buffer,
				size_t max_bytes,
				int flags,
				exs_mhandle_t mhandle);
```

| Parameter | Description |
| --------- | ----------- |
| `sockno` | specifies socket file descriptor |
| `buffer` | points to a buffer where the message should be stored |
| `max_bytes` | specifies length in bytes of the buffer point to by the buffer argument |
| `flags` | specifies additional options |
| `mhandle` | specifies a registered memory handle |

##### Test [1]: EPERM
```c
exs_blocking_recv(fd, recv_buf, RECV_BUFSIZE, flags, mh)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2]: EINVAL
```c
exs_blocking_recv(fd, recv_buf, ZERO_LENGTH, flags, mh)
```

Verify the function returns -1and sets errno to [EINVAL] if length is not at least 1.

##### Test [3]: EINVAL
```c
exs_blocking_recv(fd, recv_buf, MAX_LENGTH, flags, mh)
```

Verify the function returns -1and sets errno to [EINVAL] if length is greater than the maximum supported length for the RDMA channel.

##### Test [4]: EINVAL
```c
exs_blocking_recv(fd, BAD_BUFFER, RECV_BUFSIZE, flags, mh)
```

Verify the function returns -1and sets errno to [EINVAL] if buffer is NULL.

##### Test [5]: EOPNOTSUPP
```c
exs_blocking_recv(fd, recv_buf, RECV_BUFSIZE, BAD_FLAG, mh)
```

Verify the function returns -1and sets errno to [EOPNOTSUPP] if at least one of the flags passed to the function are not supported.
##### Test [6]: EBADF
```c
exs_blocking_recv(BAD_FD, recv_buf, RECV_BUFSIZE, flags, mh)
```

Verify the function returns -1and sets errno to [EBADF] if the function was passed a bad file descriptor.

##### Test [7]: ENOTCONN
```c
exs_blocking_recv(NOT_CONNECTED_FD, recv_buf, RECV_BUFSIZE, flags, mh)
```

Verify the function returns -1and sets errno to [ENOTCONN] if the socket was never connected to prior to the function call.

##### Test [8]: OK_EXS_BLOCKING_RECV
```c
exs_blocking_recv(fd, recv_buf, RECV_BUFSIZE, flags, mh)
```

Verify the function returns the number of bytes successfully received if passed the correct parameters.

####  exs_blocking_send()

The following tests will exercise exs_blocking_send() and its various success and error cases.

Prerequisites: exs_init(), exs_socket(), exs_connect()

```c
ssize_t exs_blocking_send(	int sockno,
				void *buffer,
				size_t max_bytes,
				int flags,
				exs_mhandle mhandle);
```

| Parameter | Description |
| --------- | ----------- |
| `sockno` | specifies socket file descriptor |
| `buffer` | points to a buffer where the message should be stored |
| `max_bytes` | specifies length in bytes of the buffer point to by the buffer argument |
| `flags` | specifies additional options |
| `mhandle` | specifies a registered memory handle |

##### Test [1]: ENOTCONN
```c
exs_blocking_send(NOT_CONNECTED_FD, send_buf, SEND_BUFSIZE, flags, mh)
```

Verify the function returns -1 and sets errno to [ENOTCONN] if the socket was never connected.

##### Test [2]: EPERM
```c
exs_blocking_send(fd, send_buf, SEND_BUFSIZE, flags, mh)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [3]: EINVAL
```c
exs_blocking_send(fd, BAD_BUFFER, SEND_BUFSIZE, flags, mh)
```

Verify the function returns -1 and sets errno to [EINVAL] if buffer is NULL.

##### Test [4]: EINVAL
```c
exs_blocking_send(fd, send_buf, SEND_BUFSIZE, EXS_UNSIGNALED | EXS_BLOCK, mh)
```

Verify the function returns -1 and sets errno to [EINVAL] if EXS_UNSIGNALED and EXS_BLOCK were both specified.

##### Test [5]: EOPNOTSUPP
```c
exs_blocking_send(fd, send_buf, SEND_BUFSIZE, BAD_FLAG, mh)
```

Verify the function returns -1 and sets errno to [EOPNOTSUPP] if flags is not supported.

##### Test [6]: EBADF
```c
exs_blocking_send(BAD_FD, send_buf, SEND_BUFSIZE, flags, mh)
```

Verify the function returns -1and sets errno to [EBADF] if the function was passed a bad file descriptor.
##### Test [8]: EBADF
```c
exs_blocking_send(fd, send_buf, SEND_BUFSIZE, flags, mh)
```

Verify the function returns -1and sets errno to [EPIPE] if the connection has been gracefully shutdown for writing.

##### Test [14] : OK_EXS_BLOCKING_SEND
```c
exs_blocking_send(fd, send_buf, SEND_BUFSIZE, flags, mh)
```

Verify the function returns the number of bytes successfully transmitted if passed the correct parameters.

Unimplemented Tests:
Verify the function returns -1 and sets errno to [ECONNRESET] if the connection has been terminated abruptly.
Verify the function returns -1 and sets errno to [EBUSY] if no send credits were available at the time of this function call.

####  exs_close()

The following tests will exercise exs_close() and its various success and error cases.

Prerequisites: exs_init(), exs_socket()

```c
int exs_close(	int sockno,
		int flags,
		exs_qhandle_t qhandle,
		exs_ahandle_t ahandle);
```

| Parameter | Description |
| --------- | ----------- |
| `sockno` | specifies the socket file descriptor |
| `flags` | specifies additional options (blocking/non-blocking) |
| `qhandle` | specifies an event queue |
| `ahandle` | specifies an arbitrary pointer value chosen by the user |

##### Test [1] : EPERM
```c
exs_close(fd, flags, qh, ah)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2] : EINVAL
```c
exs_close(fd, BAD_FLAG, qh, ah)
```

Verify the function returns -1 and sets errno to [EINVAL] if flags are not supported.

##### Test [3] : EINVAL
```c
exs_close(fd, flags, INVALID_QHANDLE, ah)
```

Verify the function returns -1 and sets errno to [EINVAL] if qhandle is invalid.

##### Test [4] : EINVAL
```c
exs_close(BAD_FD, flags, qh, ah)
```

Verify the function returns -1 and sets errno to [EBADF] if socket is invalid.

Test[5] : OK_EXS_CLOSE
```c
exs_close(fd, flags, qh, ah)
```

Verify the function returns 0 and an asynchronous event is posted to qhandle if passed the correct parameters.

####  exs_listen()

The following tests will exercise exs_listen() and its various success and error cases.

Prerequisites: exs_init(), exs_socket(), exs_bind()

```c
int exs_listen(int fd, int backlog);
```

| Parameter | Description |
| --------- | ----------- |
| `fd` | specifies socket file descriptor |
| `backlog` | specifies maximum number of outstanding client connections |

##### Test [1] : EPERM
```c
exs_listen(fd, 10)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2] : EINVAL
```c
exs_listen(fd, -1)
```

Verify the function returns -1 and sets errno to [EINVAL]if backlog is negative.

##### Test [3] : EINVAL
```c
exs_listen(CONNECTED_FD, 10)
```

Verify the function returns -1 and sets errno to [EINVAL]if fd specifies a socket that has already been connected or listened to.

##### Test [4] : EBADF
```c
exs_listen(CONNECTED_FD, 10)
```

Verify the function returns -1 and sets errno to [EBADF]if fd is invalid.

##### Test [5] : OK_EXS_LISTEN
```c
exs_listen(fd, 10)
```

Verify the function returns 0 if passed correct parameters.

####  exs_read()

The following tests will exercise exs_read() and its various success and error cases.

Prerequisites: exs_init(), exs_socket(), exs_bind(), exs_listen(), exs_accept()

```c
int exs_read(	int sockno,
		void *buffer
		size_t max_bytes);
```

| Parameter | Description |
| --------- | ----------- |
| `sockno` | specifies socket file descriptor |
| `buffer` | points to a buffer where the message should be stored |
| `max_bytes` | specifies length in bytes of the buffer point to by the buffer argument |

##### Test [1] : EPERM
```c
exs_read(fd, recv_buf, RECV_BUFSIZE)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2] : EINVAL
```c
exs_read(fd, recv_buf, ZERO_LENGTH)
```

Verify the function returns -1and sets errno to [EINVAL] if length is not at least 1.

##### Test [3] : EINVAL
```c
exs_read(fd, recv_buf, MAX_LENGTH)
```

Verify the function returns -1and sets errno to [EINVAL] if length is greater than the maximum supported length for the RDMA channel.

##### Test [4] : EINVAL
```c
exs_read(fd, INVALID_QHANDLE, RECV_BUFSIZE)
```

Verify the function returns -1and sets errno to [EINVAL] if qhandle is NULL.

##### Test [5] : EBADF
```c
exs_read(BAD_FD, recv_buf, RECV_BUFSIZE)
```

Verify the function returns -1and sets errno to [EBADF] if the function was passed a bad file descriptor.

##### Test [6] : ENOTCONN
```c
exs_read(NOT_CONNECTED_FD, recv_buf, RECV_BUFSIZE)
```

Verify the function returns -1and sets errno to [ENOTCONN] if the socket was never connected to prior to the function call.

##### Test [7] : OK_EXS_READ
```c
exs_read(fd, recv_buf, RECV_BUFSIZE)
```

Verify the function returns the number of bytes successfully received if passed the correct parameters.

####  exs_shutdown()

The following tests will exercise exs_shutdown() and its various success and error cases.

Prerequisites: exs_init(), exs_socket(), exs_connect(), exs_bind(), exs_listen(), exs_accept()

```c
int exs_shutdown(	int sockno,
			int how,
			int flags,
			exs_qhandle_t qhandle,
			exs_qhandle_t ahandle);
```

| Parameter | Description |
| --------- | ----------- |
| `sockno` | specifies socket file descriptor |
| `how` | specifies which direction the connection will be shut down |
| `flags` | additional options to be specified (blocking/non-blocking) |
| `qhandle` | specifies the destination event queue |
| `ahandle` | arbitrary pointer value chose by the user |

##### Test [1] : EPERM
```c
exs_shutdown(fd, SHUT_RDWR, flags, qh, qh)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2] : EINVAL
```c
exs_shutdown(fd, SHUT_RDWR, BAD_FLAG, qh, qh)
```

Verify the function returns -1 and sets errno to [EINVAL] if flags is invalid.

##### Test [3] : EINVAL
```c
exs_shutdown(fd, SHUT_RDWR, flags, INVALID_QHANDLE, qh)
```

Verify the function returns -1 and sets errno to [EINVAL] if qhandle is invalid.

##### Test [4] : EBADF
```c
exs_shutdown(BAD_FD, SHUT_RDWR, flags, qh, qh)
```

Verify the function returns -1 and sets errno to [EBADF] if sockno is invalid.

##### Test [5] : OK_EXS_SHUTDOWN
```c
exs_shutdown(fd, SHUT_RDWR, flags, qh, qh)
```

Verify the function returns 0 and an asynchronous event is posted to qhandle if passed the correct parameters.

####  exs_socket()

The following tests will exercise exs_socket() and its various success and error cases.

Prerequisites: exs_init()

```c
int exs_socket(	int			family,
int			socktype,
int			protocol);
```

family		must be PF_INET or PFINET6
socktype		must be SOCK_STREAM or SOCK_SEQPACKET
protocol		must be 0 or match the socktype

##### Test [1] : EPERM
```c
exs_socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [16] : OK_EXS_SOCKET
```c
exs_socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)
```

Verify the function returns a non-negative integer file descriptor fd if passed correct parameters.

##### Test [2] – [16] :
Verify the function returns -1 and sets errno to [EAFNOSUPPORT] if family is not supported.

####  exs_write()

The following tests will exercise exs_write() and its various success and error cases.

Prerequisites: exs_init(), exs_socket(), exs_connect()

```c
ssize_t exs_write(	int sockno,
			void *buffer,
			size_t max_bytes);
```

| Parameter | Description |
| --------- | ----------- |
| `sockno` | specifies socket file descriptor |
| `buffer` | points to a buffer where the message should be stored |
| `max_bytes` | specifies length in bytes of the buffer point to by the buffer argument |

##### Test [1] : ENOTCONN
```c
exs_write(NOT_CONNECTED_FD, send_buf, SEND_BUFSIZE)
```

Verify the function returns -1 and sets errno to [ENOTCONN] if the socket was never connected.

##### Test [2] : EPERM
```c
exs_write(fd, send_buf, SEND_BUFSIZE)
```

Verify the function returns -1 and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [3] : EINVAL
```c
exs_write(fd, BAD_BUFFER, SEND_BUFSIZE)
```

Verify the function returns -1 and sets errno to [EINVAL] if buffer is NULL.

##### Test [5] : EPIPE
```c
exs_write(fd, send_buf, SEND_BUFSIZE)
```

Verify the function returns -1, and sets errno to [EPIPE] if the connection has been gracefully shutdown for writing.

##### Test [11] : OK_EXS_WRITE
```c
exs_write(fd, send_buf, SEND_BUFSIZE)
```

Verify the function returns the number of bytes successfully transmitted if passed the correct parameters.

Unimplemented Tests:
Verify the function returns -1, no event is posted to qhandle, and sets errno to [ECONNRESET] if the connection has been terminated abruptly.
Verify the function returns -1, no event is posted to qhandle, and sets errno to [EINVAL] if EXS_UNSIGNALED and EXS_BLOCK were both specified.
Verify the function returns -1, no event is posted to qhandle, and sets errno to [EBUSY] if no send credits were available at the time of this function call

#### exs_blocking_sendfile()

The following tests will exercise exs_blocking_sendfile() and its various success and error cases.

Prerequisites:  exs_init(), exs_socket(), exs_connect()

```c
ssize_t exs_sendfile(	int socket,
			exs_xferfile_t *sendvec,
			int sendvec_cnt,
			int flags);
```

| Parameter | Description |
| --------- | ----------- |
| `socket` | specifies the socket file descriptor |
| `sendvec` | specifies an array of file descriptors and memory buffers |
| `sendvec_cnt` | specifies the number of elements in sendvec array |
| `flags` | specifies additional options (blocking/non-blocking) |

##### Test [1]: EPERM
```c
exs_blocking_sendfile(fd, sendvec, sendvec_cnt, flags);
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [EPERM] if exs_init() has not yet been called in the same process.

##### Test [2]: ENOTCONN
```c
exs_blocking_sendfile(NOT_CONNECTED_FD, sendvec, sendvec_cnt, flags);
```

Verify the function returns -1, no event is posted to qhandle, and sets errno to [ENOTCONN] if the socket was never connected.

##### Test [3]: EINVAL
```c
exs_blocking_sendfile(BAD_FD, sendvec, sendvec_cnt, flags);
```

Verify the function returns -1 and sets errno to [EBADF] if fd is invalid.

##### Test [4]: ENODEV
```c
exs_blocking_sendfile(fd, sendvec, sendvec_cnt, flags);
```

Verify the function returns -1 and sets errno to [ENODEV] if the file descriptor cannot be mmapped.

##### Test [5]: EINVAL
```c
exs_blocking_sendfile(fd, sendvec, sendvec_cnt, flags);
```

Verify the function returns -1 and sets errno to [EINVAL] if exs_fileds is invalid.

##### Test [6]: EINVAL
```c
exs_blocking_sendfile(fd, sendvec, sendvec_cnt, flags);
```

Verify the function returns -1 and sets errno to [EINVAL] if exs_path is invalid.

##### Test [7]: EFBIG
```c
exs_blocking_sendfile(fd, sendvec, sendvec_cnt, flags);
```

Verify the function returns -1 and sets errno to [EFBIG] if file size > 1GB.

##### Test [8]: EINVAL
```c
exs_blocking_sendfile(fd, sendvec, sendvec_cnt, flags);
```

Verify the function returns -1 and sets errno to [EINVAL] if offset + length > fize size.

##### Test [9]: EINVAL
```c
exs_blocking_sendfile(fd, sendvec, sendvec_cnt, flags);
```

Verify the function returns -1 and sets errno to [EINVAL] if offset > file size.

##### Test [10]: EINVAL
```c
exs_blocking_sendfile(fd, sendvec, BAD_SENDVEC_CNT, flags);
```

Verify the function returns -1 and sets errno to [EINVAL] if sendvec_cnt is invalid.

##### Test [11]: EINVAL
```c
exs_blocking_sendfile(fd, sendvec, sendvec_cnt, BAD_FLAG);
```

Verify the function returns -1 and sets errno to [EINVAL] if flags is invalid.

##### Test [12]: EINVAL
```c
exs_blocking_sendfile(fd, sendvec, sendvec_cnt, EXS_UNSIGNALED & EXS_BLOCK);
```

Verify the function returns -1 and sets errno to [EINVAL] if EXS_UNSIGNALED and EXS_BLOCK were both specified.

##### Test [13]: EPIPE
```c
exs_blocking_sendfile(fd, sendvec, sendvec_cnt, flags);
```

Verify the function returns -1 and sets errno to [EPIPE] if the socket is gracefully shut down for writing.


##### Test [14]: OK_EXS_BLOCKING_SENDFILE
```c
exs_blocking_sendfile(fd, sendvec, sendvec_cnt, flags);
```

Verify the function returns 0 if passed all valid parameters.

##### Test [16]: OK_EXS_BLOCKING_SENDFILE
```c
exs_blocking_sendfile(fd, sendvec, sendvec_cnt, flags);
```

Verify the function returns 0 if passed all valid parameters and exs_blocking_sendfile() is used to send the same file immediately afterward.

##### Test [15]: OK_EXS_BLOCKING_SENDFILE
```c
exs_blocking_sendfile(fd, sendvec, sendvec_cnt, flags);
```

Verify the function returns 0 if passed all valid parameters using exs_pathvec rather than exs_fdvec.
