/*
 * Copyright 2007-2017 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#include <exs.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include "inc/fault.h"

enum { msg_buf_size = 256 };
static int signalstr_max_len = 20;
static char msg_buf[msg_buf_size];

static pid_t pid_to_kill = 0;

static param *local_p;

/* handler used to catch various signals.  This handler is set with SA_RESETHAND
 * so the handler is no longer active when this function is invoked, meaning we
 * can re-raise the signal to terminate the process afterwards.
 *    also handles ALARM signal if a test times out
 *    */
static void fault_handler(int signo)
{
	char *signalstr;
	int ret, exit_code;

	switch (signo) {
	case SIGABRT:
		signalstr = "Aborted\n";
		break;
	case SIGALRM:
		signalstr = "Timed Out\n";
		break;
	case SIGSEGV:
		signalstr = "Segmentation Fault\n";
		break;
	default:
		signalstr = "Unexpected Signal\n";
		break;
	}
	strncat(msg_buf, signalstr, signalstr_max_len);
	exit_code = EXIT_FAILURE;

	/* If we can't print this error, there's not much we can do
	 * here :( */
	ret = write(STDOUT_FILENO, msg_buf, strnlen(msg_buf, msg_buf_size));
	if (ret < 256) {
		exit_code = 2;
	}

	if (pid_to_kill != 0) {
		kill(pid_to_kill, SIGTERM);
	}

	/* Re-raise the signal; without this handler in place (thanks to
	 * SA_RESETHAND), the signal will kill the process and let our parent
	 * know what signal we got */
	if (kill(getpid(), signo)) {
		_exit(exit_code);
	}
}				/* fault_handler */

/* initializes fault handler */
int setup_fault_catcher(param *p, pid_t kill_this_pid)
{
	struct sigaction act_params;
	int ret;

	local_p = p;
	pid_to_kill = kill_this_pid;

	act_params.sa_handler = fault_handler;
	sigfillset(&act_params.sa_mask);
	act_params.sa_flags = SA_RESETHAND|SA_NODEFER;

	ret = snprintf(msg_buf, msg_buf_size,
		"%-2d : %-23s : %-100s : %s : ",
		local_p->test_number, local_p->test_value,
		local_p->test_name, "FAIL");
	assert(ret > 0 && ret < msg_buf_size - signalstr_max_len);

	/* set up handler for "Segmentation violation" */
	if (sigaction(SIGSEGV, &act_params, NULL) != 0) {
		perror("sigaction SIGSEGV");
		exs_close(p->fd, EXS_BLOCK | EXS_DONTLINGER, NULL, NULL);
		return -1;
	}
	/* set up handler for "Abort" */
	if (sigaction(SIGABRT, &act_params, NULL) != 0) {
		perror("sigaction SIGABRT");
		exs_close(p->fd, EXS_BLOCK | EXS_DONTLINGER, NULL, NULL);
		return -1;
	}
	/* set up handler for "Alarm clock" */
	if (sigaction(SIGALRM, &act_params, NULL) != 0) {
		fprintf(stdout, "ALARMAMAMAM");
		perror("sigaction SIGALRM");
		exs_close(p->fd, EXS_BLOCK | EXS_DONTLINGER, NULL, NULL);
		return -1;
	}

	return 0;
} /* setup_fault_catcher */
