/*
 * Copyright 2007-2017 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#include "result.h"
#include "setup/exs_init_setup.h"
#include "setup/exs_socket_setup.h"
#include "setup/exs_connect_setup.h"
#include "setup/exs_bind_setup.h"
#include "setup/exs_listen_setup.h"
#include "setup/exs_accept_setup.h"
#include "setup/exs_client_setup.h"

#define BAD_FLAG 2

#define ACCEPT_TESTS 8

enum get_bound_socket_flags {
	SOCKET_LISTEN = 0x1,
	SOCKET_NOBUFS = 0x2,
	SOCKET_ALL_FLAGS = 0x3,
};

static char *all_accept_tests[ACCEPT_TESTS] = {
	"exs_accept(fd, &acceptaddr, 1, flags, qh) [no exs_init]",
	"exs_accept(BAD_FD, &acceptaddr, 1, flags, qh)",
	"exs_accept(fd, &acceptaddr, 0, flags, qh) [no exs_listen]",
	"exs_accept(fd, &acceptaddr, NEGATIVE_ADDRVEC_CNT, 0, qh) [addrvect <= when non blocking]",
	"exs_accept(fd, &acceptaddr, ZERO_ADDRVEC_CNT, EXS_BLOCK, qh) [addrvec_cnet != 1 when blocking]",
	"exs_accept(fd, &acceptaddr, 1, BAD_FLAG, qh)",
	"exs_accept(fd, &acceptaddr, 1, flags, INVALID_QHANDLE)",
	"exs_accept(fd, &acceptaddr, 1, flags, qh)"
};

static void *test_accept(void *test_void_ptr)
{
	param *p = (param *) test_void_ptr;
	int test = p->test_number;

	struct addrinfo *info;
	struct addrinfo hints;
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = p->family;
	hints.ai_socktype = p->socktype;
	hints.ai_protocol = p->protocol;
	hints.ai_flags = AI_PASSIVE;

	int r = getaddrinfo(p->addr, p->port, &hints, &info);
	if (r < 0) {
		print_error("getaddrinfo");
		exit(EXIT_FAILURE);
	}

	struct sockaddr_storage addr_store;
	exs_acceptaddr_t acceptaddr;

	memset(&acceptaddr, 0, sizeof(acceptaddr));
	acceptaddr.exs_addr = (struct sockaddr *)&addr_store;
	acceptaddr.exs_addrlen = sizeof(addr_store);

	/* EPERM: exs_init has not been called */
	if (test == 1) {
		p->test_value = "EPERM";
		p->test_name = all_accept_tests[test - 1];
		setup_fault_catcher(p, 0);
		int ret =
		    exs_accept(p->fd, &acceptaddr, 1, p->flags, p->qh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EPERM);
	}

	/* EBADF: socketno is not a valid file descriptor */
	else if (test == 2) {
		p->test_value = "EBADF";
		p->test_name = all_accept_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_bind(p) < 0)
			exit(EXIT_FAILURE);

		int ret = exs_accept(-1, &acceptaddr, 1, p->flags, p->qh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EBADF);
	}

	/* EINVAL: exs_listen has not been called */
	else if (test == 3) {
		p->test_value = "EINVAL";
		p->test_name = all_accept_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_bind(p) < 0)
			exit(EXIT_FAILURE);
		int ret =
		    exs_accept(p->fd, &acceptaddr, 0, p->flags, p->qh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* EINVAL: addrvect <=0 when non blocking */
	else if (test == 4) {
		p->test_value = "EINVAL";
		p->test_name = all_accept_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		int ret = exs_accept(p->fd, &acceptaddr, -1, 0, p->qh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* EINVAL: addrvec_cnet != 1 when blocking */
	else if (test == 5) {
		p->test_value = "EINVAL";
		p->test_name = all_accept_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		int ret =
		    exs_accept(p->fd, &acceptaddr, 0, p->flags, p->qh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* EINVAL: flag not supported */
	else if (test == 6) {
		p->test_value = "EINVAL";
		p->test_name = all_accept_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		int ret =
		    exs_accept(p->fd, &acceptaddr, 1, BAD_FLAG, p->qh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* EINVAL: qhandle is not valid */
	else if (test == 7) {
		p->test_value = "EINVAL";
		p->test_name = all_accept_tests[test - 1];
		setup_fault_catcher(p, 0);
		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_bind(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_listen(p) < 0)
			exit(EXIT_FAILURE);
		int ret = exs_accept(p->fd, &acceptaddr, 1, 0, p->qh);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* OK_EXS_ACCEPT */
	else if (test == 8) {
		pid_t pid = fork();
		if (pid == 0) {
			exs_client(p);
			exit(EXIT_SUCCESS);
		} else {
			setup_fault_catcher(p, pid);
			p->test_value = "OK_EXS_ACCEPT";
			p->test_name = all_accept_tests[test - 1];
			int ret, fd = -1;

			if (ok_exs_init(p) < 0)
				exit(EXIT_FAILURE);

			if (p->mode > 0) {
				if (ok_exs_qcreate(p) < 0)
					exit(EXIT_FAILURE);
			}

			if (ok_exs_socket(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_bind(p) < 0)
				exit(EXIT_FAILURE);
			if (ok_exs_listen(p) < 0)
				exit(EXIT_FAILURE);

			if (p->mode == 0) {
				fd = open(exs_test_fifo_1, O_WRONLY);
			}

			ret = ok_exs_accept(p);

			if (p->mode > 0) {
				if (ok_exs_qdequeue(p) < 0)
					exit(EXIT_FAILURE);
			}

			if (ret == 0) {
				print_macro(test, p->test_value,
					    p->test_name, ret, 0, 0);
			} else {
				print_macro(test, p->test_value,
					    p->test_name, ret, -1, EINVAL);
			}

			if (p->mode == 0) {
				close(fd);
				exs_close(p->fd, p->close_flags, p->qh,
					  p->ah);
				fd = open(exs_test_fifo_2, O_RDONLY);
				close(fd);
				exit(EXIT_SUCCESS);
			}
		}
	}

	/** test number not valid **/
	else if (test != 0) {
		fprintf(stdout, "Invalid Test Number: %d\n", test);
	}

	return NULL;
}

int main(int argc, char *argv[])
{
	execute(argc, argv, test_accept, all_accept_tests, ACCEPT_TESTS);
}
