/*
 * Copyright 2007-2017 University of New Hampshire Interoperability Laboratory
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grants No. OCI-1127228 and DGE-0913620.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#include "result.h"
#include "setup/exs_bind_setup.h"
#include "setup/exs_init_setup.h"
#include "setup/exs_socket_setup.h"

#define BAD_FLAG 2

#define BIND_TESTS 4

static char *all_bind_tests[BIND_TESTS] = {
	"exs_bind(fd,ai_addr,ai_addrlen) [no exs_init]",
	"exs_bind(BAD_FD,ai_addr,ai_addrlen)",
	"exs_bind(CONNECTED_FD,ai_addr,ai_addrlen)",
	"exs_bind(fd,ai_addr,ai_addrlen)"
};

static void *test_bind(void *test_void_ptr)
{
	param *p = (param *) test_void_ptr;
	int test = p->test_number;

	struct addrinfo *info;
	struct addrinfo hints;
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = p->family;
	hints.ai_socktype = p->socktype;
	hints.ai_protocol = p->protocol;
	hints.ai_flags = AI_PASSIVE;

	int r = getaddrinfo(p->addr, p->port, &hints, &info);
	if (r < 0) {
		print_error("getaddrinfo");
		exit(EXIT_FAILURE);
	}

	/* EPERM: exs_init has not been called */
	if (test == 1) {
		p->test_value = "EPERM";
		p->test_name = all_bind_tests[test - 1];
		setup_fault_catcher(p, 0);

		int ret = exs_bind(p->fd, info->ai_addr, info->ai_addrlen);
		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EPERM);
	}

	/* EBADF: fd is invalid */
	else if (test == 2) {
		p->test_value = "EBADF";
		p->test_name = all_bind_tests[test - 1];
		setup_fault_catcher(p, 0);

		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);

		int ret = exs_bind(-1, info->ai_addr, info->ai_addrlen);

		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EBADF);
	}

	/* ENIVAL: socket provided by fd has already been bound */
	else if (test == 3) {
		p->test_value = "EINVAL";
		p->test_name = all_bind_tests[test - 1];
		setup_fault_catcher(p, 0);

		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_bind(p) < 0)
			exit(EXIT_FAILURE);

		int ret = exs_bind(p->fd, info->ai_addr, info->ai_addrlen);

		print_macro(test, p->test_value, p->test_name, ret, -1,
			    EINVAL);
	}

	/* OK_EXS_BIND */
	else if (test == 4) {
		p->test_value = "EINVAL";
		p->test_name = all_bind_tests[test - 1];
		setup_fault_catcher(p, 0);

		if (ok_exs_init(p) < 0)
			exit(EXIT_FAILURE);
		if (ok_exs_socket(p) < 0)
			exit(EXIT_FAILURE);

		int ret = ok_exs_bind(p);

		if (ret == 0) {
			print_macro(test, p->test_value, p->test_name, ret,
				    0, 0);
		} else {
			print_macro(test, p->test_value, p->test_name, ret,
				    -1, ENOTSUP);
		}
	}

	/** test number not valid **/
	else if (test != 0) {
		fprintf(stdout, "Invalid Test Number: %d\n", test);
	}

	return NULL;
}

int main(int argc, char *argv[])
{				//arv[1] first arg from test_all.sh
	execute(argc, argv, test_bind, all_bind_tests, BIND_TESTS);
}
